# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from talos.thrift.Thrift import TType

TYPE_IDX = 1
SPEC_ARGS_IDX = 3
SPEC_ARGS_CLASS_REF_IDX = 0
SPEC_ARGS_THRIFT_SPEC_IDX = 1


def fix_spec(all_structs):
    """Wire up recursive references for all TStruct definitions inside of each thrift_spec."""
    for struc in all_structs:
        spec = struc.thrift_spec
        for thrift_spec in spec:
            if thrift_spec is None:
                continue
            elif thrift_spec[TYPE_IDX] == TType.STRUCT:
                other = thrift_spec[SPEC_ARGS_IDX][SPEC_ARGS_CLASS_REF_IDX].thrift_spec
                thrift_spec[SPEC_ARGS_IDX][SPEC_ARGS_THRIFT_SPEC_IDX] = other
            elif thrift_spec[TYPE_IDX] in (TType.LIST, TType.SET):
                _fix_list_or_set(thrift_spec[SPEC_ARGS_IDX])
            elif thrift_spec[TYPE_IDX] == TType.MAP:
                _fix_map(thrift_spec[SPEC_ARGS_IDX])


def _fix_list_or_set(element_type):
    # For a list or set, the thrift_spec entry looks like,
    # (1, TType.LIST, 'lister', (TType.STRUCT, [RecList, None], False), None, ),  # 1
    # so ``element_type`` will be,
    # (TType.STRUCT, [RecList, None], False)
    if element_type[0] == TType.STRUCT:
        element_type[1][1] = element_type[1][0].thrift_spec
    elif element_type[0] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[1])
    elif element_type[0] == TType.MAP:
        _fix_map(element_type[1])


def _fix_map(element_type):
    # For a map of key -> value type, ``element_type`` will be,
    # (TType.I16, None, TType.STRUCT, [RecMapBasic, None], False), None, )
    # which is just a normal struct definition.
    #
    # For a map of key -> list / set, ``element_type`` will be,
    # (TType.I16, None, TType.LIST, (TType.STRUCT, [RecMapList, None], False), False)
    # and we need to process the 3rd element as a list.
    #
    # For a map of key -> map, ``element_type`` will be,
    # (TType.I16, None, TType.MAP, (TType.I16, None, TType.STRUCT,
    #  [RecMapMap, None], False), False)
    # and need to process 3rd element as a map.

    # Is the map key a struct?
    if element_type[0] == TType.STRUCT:
        element_type[1][1] = element_type[1][0].thrift_spec
    elif element_type[0] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[1])
    elif element_type[0] == TType.MAP:
        _fix_map(element_type[1])

    # Is the map value a struct?
    if element_type[2] == TType.STRUCT:
        element_type[3][1] = element_type[3][0].thrift_spec
    elif element_type[2] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[3])
    elif element_type[2] == TType.MAP:
        _fix_map(element_type[3])



#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from talos.thrift.Thrift import TType, TMessageType, TException, TApplicationException
from talos.thrift.auth.ttypes import *

SIGNATURE_SUPPORT = {
    1 : False,
    2 : True,
    10 : True,
    11 : True,
    12 : False,
    13 : False,
}
HK_HOST = "Host"
HK_TIMESTAMP = "X-Xiaomi-Timestamp"
MI_DATE = "x-xiaomi-date"
XIAOMI_HEADER_PREFIX = "x-xiaomi-"
HK_CONTENT_MD5 = "X-Xiaomi-Content-MD5"
HK_AUTHORIZATION = "Authorization"
HK_SERVICE_ADMIN = "Service-Admin"
HK_SERVICE_MARK = "#"
HK_ATTACHED_INFO = "Attached-Info"
SUGGESTED_SIGNATURE_HEADERS = [
  "Host",
  "X-Xiaomi-Timestamp",
  "X-Xiaomi-Content-MD5",
]



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.thrift.common.ttypes import ErrorCode
from talos.thrift.common.ttypes import HttpStatusCode


class TalosTransportError(Exception):
	HttpStatusCode = HttpStatusCode
	ErrorCode = int
	ErrorMessage = str
	ServerTime = int

	def __init__(self, httpStatusCode=None, errorMessage=None, timestamp=None):
		errorCode = None
		if httpStatusCode == HttpStatusCode.INVALID_AUTH:
			errorCode = ErrorCode.INVALID_AUTH_INFO
		elif httpStatusCode == HttpStatusCode.CLOCK_TOO_SKEWED:
			errorCode = ErrorCode.CLOCK_TOO_SKEWED
		elif httpStatusCode == HttpStatusCode.REQUEST_TOO_LARGE:
			errorCode = ErrorCode.REQUEST_TOO_LARGE
		elif httpStatusCode == HttpStatusCode.INTERNAL_ERROR:
			errorCode = ErrorCode.INTERNAL_SERVER_ERROR
		elif httpStatusCode == HttpStatusCode.BAD_REQUEST:
			errorCode = ErrorCode.BAD_REQUEST
		self.HttpStatusCode = httpStatusCode
		self.ErrorCode = errorCode
		self.ErrorMessage = errorMessage
		self.ServerTime = timestamp

	def get_error_code(self):
		return self.ErrorCode

	def string(self):
		if not self:
			return None
		errorString = "Transport error, ErrorCode: " + str(self.ErrorCode) + \
					  ", HttpStatusCode: " + str(self.HttpStatusCode) + \
					  ", ErrorMessage: " + self.ErrorMessage
		return errorString

	def error(self):
		return self.string()


class InvalidArgumentError(Exception):
	errorMsg = str

	def __init__(self, errorMsg=None):
		self.errorMsg = errorMsg

	def to_string(self):
		return self.errorMsg




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.thrift.topic.ttypes import GetTopicAttributeRequest
from talos.thrift.topic.ttypes import GetTopicAttributeResponse
from talos.thrift.message.ttypes import MessageType
from talos.client.Constants import Constants
from atomic import AtomicLong
from talos.producer.BufferedMessageCount import BufferedMessageCount
from talos.producer.UserMessage import UserMessage
from talos.producer.ProducerNotActiveException import ProducerNotActiveException
from talos.client.TalosClientFactory import TalosClientFactory
from talos.client.TalosErrors import InvalidArgumentError
from talos.admin.TalosAdmin import TalosAdmin
from talos.utils import Utils
from talos.utils.Utils import synchronized
from talos.producer.PartitionSender import PartitionSender
from talos.thrift.common.ttypes import GalaxyTalosException
from threading import Timer
import threading
import random
import logging
import traceback


class PRODUCER_STATE:
    ACTIVE = 0
    DISABLED = 1
    SHUTDOWN = 2


class TalosProducer(object):
    logger = logging.getLogger("TalosProducer")
    partitionKeyMinLen = Constants.TALOS_PARTITION_KEY_LENGTH_MINIMAL
    partitionKeyMaxLen = Constants.TALOS_PARTITION_KEY_LENGTH_MAXIMAL
    requestId = AtomicLong(1)
    globalLock = threading.Condition()
    partitionNumber = int
    curPartitionId = int

    messageCallbackExecutors = threading.Thread
    partitionCheckExecutor = threading.Timer
    partitionSenderMap = dict()

    lock = threading.Lock()

    def __init__(self, producerConfig=None, credential=None, topicName=None, talosAdmin=None,
                 talosClientFactory=None, topicTalosResourceName=None, partitioner=None,
                 topicAbnormalCallback=None, userMessageCallback=None):
        self.talosProducerConfig = producerConfig
        self.producerState = PRODUCER_STATE.ACTIVE
        if talosClientFactory:
            self.talosClientFactory = talosClientFactory
        else:
            self.talosClientFactory = TalosClientFactory(self.talosProducerConfig, credential)
        if talosAdmin:
            self.talosAdmin = talosAdmin
        else:
            self.talosAdmin = TalosAdmin(self.talosClientFactory)
        self.topicName = topicName
        if topicTalosResourceName:
            self.topicTalosResourceName = topicTalosResourceName
        else:
            self.topicTalosResourceName = self.talosAdmin.get_topic_attribute(
                GetTopicAttributeRequest(topicName=self.topicName))
        self.partitioner = partitioner
        self.topicAbnormalCallback = topicAbnormalCallback
        self.userMessageCallback = userMessageCallback
        self.updatePartitionIdInterval = producerConfig.get_check_partition_interval()
        self.lastUpdatePartitionIdTime = Utils.current_time_mills()
        self.updatePartitionIdMsgNumber = producerConfig.get_update_partition_msg_number()
        self.lastAddMsgNumber = 0
        self.maxBufferedMsgTime = producerConfig.get_max_buffered_msg_time()
        self.maxBufferedMsgNumber = producerConfig.get_max_buffered_msg_number()
        self.maxBufferedMsgBytes = producerConfig.get_max_buffered_msg_bytes()
        self.bufferedCount = BufferedMessageCount(self.maxBufferedMsgNumber,
                                                  self.maxBufferedMsgBytes)
        self.clientId = Utils.generate_client_id(producerConfig.get_client_ip(), "producer")
        self.synchronizedLock = threading.Lock()

        # getTopicTalosResourceName(producerConfig, credential);
        self.check_and_get_topic_info(self.topicTalosResourceName.topicTalosResourceName)

        self.init_partition_sender()
        self.init_check_partition_task()

        self.logger.info("Init a producer for topic: " +
                         self.topicTalosResourceName.topicTalosResourceName +
                         ", partitions: " + str(self.partitionNumber))

    def check_partition_task(self):
        response = GetTopicAttributeResponse()
        try:
            response = self.talosAdmin.get_topic_attribute(
                GetTopicAttributeRequest(topicName=self.topicName))
        except Exception as e:
            if isinstance(e, GalaxyTalosException):
                self.logger.error(e.errMsg)
            self.logger.error("Exception in CheckPartitionTask: "
                              + str(traceback.format_exc()))
            if Utils.is_topic_not_exist(e):
                self.disable_producer(e)
            return

        if not self.topicTalosResourceName == response.topicTalosResourceName:
            errMsg = "The topic: " + self.topicTalosResourceName.topicTalosResourceName + \
                     " not exist. It might have been deleted. The putMessage threads" + \
                     " will be cancel."

            self.logger.error(errMsg)
            # cancel the putMessage thread
            self.disable_producer(Exception(errMsg))
            return

        topicPartitionNum = response.topicAttribute.partitionNumber
        if not self.partitionNumber == topicPartitionNum:
            self.logger.info("Adjust partitionSender and partitionNumber from: " +
                             str(self.partitionNumber) + " to: " + str(topicPartitionNum))
            # adjust partitionSender(allow decreasing)
            self.adjust_partition_sender(topicPartitionNum)
            # update partitionNumber
            self.set_partition_number(topicPartitionNum)

        self.partitionCheckExecutor = threading.Timer(interval=self.talosProducerConfig.
                                                      get_producer_check_partition_interval(),
                                                      function=self.check_partition_task)
        self.partitionCheckExecutor.setName("partitionCheckTask-" + self.topicName)
        self.partitionCheckExecutor.start()

    @synchronized
    def add_user_message(self, msgList=None, partitionId=None):
        if not self.is_active():
            raise ProducerNotActiveException("Producer is not active, current state: "
                                             + str(self.producerState))

        # check total buffered message number
        while self.bufferedCount.is_full():
            with self.lock:
                try:
                    self.logger.info("too many buffered messages, globalLock is active."
                                     + " message number: " +
                                     str(self.bufferedCount.get_buffered_msg_number())
                                     + ", message bytes:  " +
                                     str(self.bufferedCount.get_buffered_msg_bytes()))
                    # release global lock but no notify
                    self.globalLock.acquire()
                    self.globalLock.wait()
                    self.globalLock.release()
                except Exception as e:
                    if isinstance(e, GalaxyTalosException):
                        self.logger.error(e.errMsg)
                    self.logger.error("addUserMessage global lock wait is interrupt."
                                      + str(traceback.format_exc()))

        if not partitionId:
            self.do_add_user_message(msgList)
        else:
            userMessageList = []

            for message in msgList:
                # set timestamp and messageType if not set
                Utils.update_message(message, MessageType.BINARY)
                # check data validity
                Utils.check_message_validity(message)
                userMessageList.append(UserMessage(message))

            assert partitionId in self.partitionSenderMap
            self.partitionSenderMap.get(partitionId).add_message(userMessageList)

    # @synchronized
    def do_add_user_message(self, msgList=None):
        # user can optionally set 'partitionKey' and 'sequenceNumber' when construct Message
        partitionBufferMap = dict()

        # check / update curPartitionId
        if self.should_update_partition():
            self.curPartitionId = (self.curPartitionId + 1) % self.partitionNumber
            self.lastUpdatePartitionIdTime = Utils.current_time_mills()
            self.lastAddMsgNumber = 0
        currentPartitionId = self.curPartitionId
        self.lastAddMsgNumber += len(msgList)

        partitionBufferMap[currentPartitionId] = []

        for message in msgList:
            # set timestamp and messageType if not set;
            Utils.update_message(message, MessageType.BINARY)
            # check data validity
            Utils.check_message_validity(message)

            # check partitionKey setting and validity
            if not message.partitionKey:
            # straight forward put to cur partitionId queue
                userMessage = UserMessage(message)
                partitionBufferMap.get(currentPartitionId).append(userMessage)
            else:
                self.check_message_partition_key_validity(message.getPartitionKey())
                # construct UserMessage and dispatch to buffer by partitionId
                partitionId = self.get_partition_id(message.partitionKey)
                if partitionId not in partitionBufferMap:
                    partitionBufferMap[partitionId] = []
                    userMessage = UserMessage(message)
                    partitionBufferMap.get(partitionId).append(userMessage)

        # add to partitionSender
        for key, value in partitionBufferMap.items():
            partitionId = key
            assert partitionId in self.partitionSenderMap
            self.partitionSenderMap.get(partitionId).add_message(value)

    def should_update_partition(self):
        return (Utils.current_time_mills() - self.lastUpdatePartitionIdTime >=
                self.updatePartitionIdInterval) or (self.lastAddMsgNumber >=
                                                    self.updatePartitionIdMsgNumber)

    def check_message_partition_key_validity(self, partitionKey=None):
        assert not partitionKey
        if len(partitionKey) < self.partitionKeyMinLen or len(partitionKey) > self.partitionKeyMaxLen:
            raise InvalidArgumentError("Invalid partition key which length must be at "
                                       + "least " + str(self.partitionKeyMinLen) +
                                       " and at most " + str(self.partitionKeyMinLen) +
                                       ", got " + partitionKey.length())

    def get_partition_id(self, partitionKey=None):
        return self.partitioner.partition(partitionKey, self.partitionNumber)

    # cancel the putMessage threads and checkPartitionTask
    # when topic not exist during producer running
    def disable_producer(self, exception=None):
        if not self.is_active():
            return

        self.producerState = PRODUCER_STATE.DISABLED
        self.stop_and_wait()
        self.topicAbnormalCallback.abnormal_handler(self.topicTalosResourceName, exception)

    def is_active(self):
        return self.producerState == PRODUCER_STATE.ACTIVE

    def is_disabled(self):
        return self.producerState == PRODUCER_STATE.DISABLED

    def is_shutdown(self):
        return self.producerState == PRODUCER_STATE.SHUTDOWN

    def shutdown(self):
        if not self.is_active():
            return
        self.producerState = PRODUCER_STATE.SHUTDOWN
        self.stop_and_wait()

    def stop_and_wait(self):
        for key, value in self.partitionSenderMap.items():
            value.shutdown()
        self.partitionCheckExecutor.cancel()

    def adjust_partition_sender(self, newPartitionNum=None):
        if self.partitionNumber < newPartitionNum:
            partitionId = self.partitionNumber
            while partitionId < newPartitionNum:
                self.create_partition_sender(partitionId)
                partitionId += 1
        elif self.partitionNumber > newPartitionNum:
            partitionId = newPartitionNum
            while partitionId < self.partitionNumber:
                self.release_partition_sender(partitionId)
                partitionId += 1

    @synchronized
    def create_partition_sender(self, partitionId=None):
        partitionSender = PartitionSender(partitionId=partitionId,
                                          topicName=self.topicName,
                                          topicTalosResourceName=
                                          self.topicTalosResourceName,
                                          requestId=self.requestId, clientId=self.clientId,
                                          talosProducerConfig=self.talosProducerConfig,
                                          messageClient=self.talosClientFactory.
                                          new_message_client(), userMessageCallback=
                                          self.userMessageCallback,
                                          messageCallbackExecutors=
                                          self.messageCallbackExecutors,
                                          globalLock=self.globalLock, producer=self)
        self.partitionSenderMap[partitionId] = partitionSender

    @synchronized
    def release_partition_sender(self, partitionId=None):
        self.partitionSenderMap.get(partitionId).close()
        del self.partitionSenderMap[partitionId]

    def set_partition_number(self, partitionNum=None):
        self.partitionNumber = partitionNum

    def check_and_get_topic_info(self, topicTalosResourceName=None):
        topicName = Utils.get_topic_name_by_resource_name(topicTalosResourceName.topicTalosResourceName)
        response = self.talosAdmin.get_topic_attribute(
            GetTopicAttributeRequest(topicName))

        if not topicTalosResourceName == response.topicTalosResourceName:
            raise InvalidArgumentError("The topic: " +
                                       topicTalosResourceName.topicTalosResourceName
                                       + " not found")

        self.partitionNumber = response.topicAttribute.partitionNumber
        self.curPartitionId = random.randint(0, self.partitionNumber)
        self.topicTalosResourceName = topicTalosResourceName

    def init_partition_sender(self):
        partitionId = 0
        while partitionId < self.partitionNumber:
            self.create_partition_sender(partitionId)
            partitionId += 1

    @synchronized
    def init_check_partition_task(self):
        self.partitionCheckExecutor = Timer(interval=self.talosProducerConfig.
                                            get_producer_check_partition_interval(),
                                            function=self.check_partition_task)
        self.partitionCheckExecutor.setName("partitionCheckTask-" + self.topicName)
        self.partitionCheckExecutor.start()

    def increase_buffered_count(self, incrementNumber=None, incrementBytes=None):
        self.bufferedCount.increase(incrementNumber, incrementBytes)

    def decrease_buffered_count(self, decrementNumber=None, decrementBytes=None):
        self.bufferedCount.decrease(decrementNumber, decrementBytes)





#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.thrift.auth.ttypes import Credential
from talos.thrift.auth.ttypes import UserType
from talos.thrift.message.ttypes import Message
from talos.client.TalosClientConfig import TalosClientConfig
from talos.producer.SimpleProducer import SimpleProducer
from atomic import AtomicLong
import time
import logging
import traceback


class SimpleProducerDemo:
    logger = logging.getLogger("SimpleProducerDemo")
    accessKey = "$yourAccessKey"
    accessSecret = "$yourSecretKey"
    topicName = "$yourTopicName"
    partitionId = 0

    pro = dict()
    pro["galaxy.talos.service.endpoint"] = "$yourEndpoint"
    clientConfig = TalosClientConfig
    credential = Credential
    successPutNumber = AtomicLong(0)

    simpleProducer = SimpleProducer

    def __init__(self):
        self.clientConfig = TalosClientConfig(self.pro)
        # credential
        self.credential = Credential(UserType.DEV_XIAOMI, self.accessKey, self.accessSecret)

    def start(self):
        # init producer
        self.simpleProducer = SimpleProducer(producerConfig=self.clientConfig,
                                             topicName=self.topicName,
                                             partitionId=self.partitionId,
                                             credential=self.credential)
        message = Message()
        message.message = "test message: this message is a text string."

        messageList = [message]
        # a toy demo for putting messages to Talos server continuously
        # by using a infinite loop
        while True:
            try:
                self.simpleProducer.put_message_list(messageList)
                time.sleep(2)
            except Exception as e:
                self.logger.warn("put message failed, try again", traceback.print_exc())
                raise e
            successNum = self.successPutNumber.value
            self.successPutNumber.get_and_set(successNum + 1)
            self.logger.info("success put message count: " +
                             str(self.successPutNumber.value))
            print("success put message count: " + str(self.successPutNumber.value))


simpleProducerDemo = SimpleProducerDemo()
# add message list to producer continuously
simpleProducerDemo.start()



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#
from talos.consumer.TalosMessageReader import TalosMessageReader


class MessageReaderFactory:
    def create_message_reader(self, talosConsumerConfig=None):
        return TalosMessageReader(talosConsumerConfig)




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.consumer.MessageProcessor import MessageProcessor
from talos.consumer.SimpleConsumer import SimpleConsumer
from talos.thrift.topic.ttypes import TopicAndPartition
from talos.thrift.message.ttypes import MessageOffset
from talos.client.TalosClientFactory import ConsumerClient
from talos.client.TalosClientFactory import MessageClient
from talos.client.TalosClientConfig import TalosClientConfig
from atomic import AtomicLong
from talos.utils import Utils
import traceback
import logging
import abc
import time


class MessageReader:
    logger = logging.getLogger("MessageReader")
    commitThreshold = int
    commitInterval = int
    fetchInterval = int

    lastCommitTime = int
    lastFetchTime = int

    startOffset = AtomicLong
    finishedOffset = int
    lastCommitOffset = int
    messageProcessor = MessageProcessor

    workerId = str
    consumerGroup = str
    topicAndPartition = TopicAndPartition
    consumerConfig = TalosClientConfig
    simpleConsumer = SimpleConsumer
    messageClient = MessageClient
    consumerClient = ConsumerClient
    queryOffsetClient = ConsumerClient

    # outer - checkPoint can be used only one time, burn after reading to prevent
    # partitionFetcher re - lock() and re - use outer - checkPoint again by consumer re - balance
    outerCheckPoint = int

    def __init__(self, consumerConfig=None):
        self.consumerConfig = consumerConfig
        self.lastCommitOffset = self.finishedOffset = -1
        self.lastCommitTime = self.lastFetchTime = Utils.current_time_mills()
        self.startOffset = AtomicLong(-1)
        if consumerConfig:
            self.commitThreshold = consumerConfig.get_commit_offset_threshold()
            self.commitInterval = consumerConfig.get_commit_offset_interval()
            self.fetchInterval = consumerConfig.get_fetch_message_interval()
        self.outerCheckPoint = None

    def set_worker_id(self, workerId=None):
        self.workerId = workerId

    def set_consumer_group(self, consumerGroup=None):
        self.consumerGroup = consumerGroup

    def set_topic_and_partition(self, topicAndPartition=None):
        self.topicAndPartition = topicAndPartition

    def set_simple_consumer(self, simpleConsumer=None):
        self.simpleConsumer = simpleConsumer

    def set_message_processor(self, messageProcessor=None):
        self.messageProcessor = messageProcessor

    def set_message_client(self, messageClient=None):
        self.messageClient = messageClient

    def set_consumer_client(self, consumerClient=None):
        self.consumerClient = consumerClient

    def set_query_offset_client(self, queryOffsetClient=None):
        self.queryOffsetClient = queryOffsetClient

    def set_outer_checkpoint(self, outerCheckpoint=None):
        self.outerCheckPoint = outerCheckpoint

    def get_start_offset(self):
        return self.startOffset.value

    def get_cur_checkpoint(self):
        # init state or before the first committing
        if self.lastCommitOffset <= self.startOffset.value:
            return self.startOffset.value

        # from lastCommitOffset + 1 when reading next time
        return self.lastCommitOffset + 1

    def should_commit(self, isContinuous=None):
        interval = Utils.current_time_mills() - self.lastCommitTime
        if isContinuous:
            return interval >= self.commitInterval or self.finishedOffset - self.lastCommitOffset >= self.commitThreshold
        else:
            return interval >= self.commitInterval and self.finishedOffset - self.lastCommitOffset >= self.commitThreshold

    def clean_reader(self):
        # wait task quit gracefully: stop reading, commit offset, clean and shutdown
        if self.finishedOffset > self.lastCommitOffset:
            try:
                self.commit_check_point()
            except Exception as e:
                self.logger.error("Error when commit offset for topic: " +
                                  str(self.topicAndPartition.topicTalosResourceName) +
                                  " partition: " + str(self.topicAndPartition.partitionId)
                                  + str(traceback.format_exc()))

    def process_fetch_exception(self, e=None):
        # delay when partitionNotServing
        if Utils.is_partition_not_serving(e):
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug("Partition: " + str(self.topicAndPartition.partitionId) +
                                  " is not serving state, sleep a while for waiting it work."
                                  + str(traceback.format_exc()))
            time.sleep(self.consumerConfig.get_wait_partition_working_time())

        if Utils.is_partition_not_exist(e):
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug("Partition: " + str(self.topicAndPartition.partitionId) +
                                  " not exist anymore, sleep a while for waiting it reduce."
                                  + str(traceback.format_exc()))
            time.sleep(self.consumerConfig.get_wait_partition_reduce_time())

        # if process message offset out of range, reset start offset
        if Utils.is_offset_out_of_range(e):
            if self.consumerConfig.is_reset_latest_offset_out_of_range():
                self.logger.warning("Got PartitionOutOfRange error, " +
                                    " offset by current latest offset" +
                                    str(traceback.format_exc()))
                self.startOffset.get_and_set(MessageOffset.LATEST_OFFSET)
                self.lastCommitOffset = self.finishedOffset = - 1
                self.lastCommitTime = Utils.current_time_mills()
            else:
                self.logger.warning("Got PartitionOutOfRange error," +
                                    " reset offset by current start offset" +
                                    str(traceback.format_exc()))
                self.startOffset.get_and_set(MessageOffset.START_OFFSET)
                self.lastCommitOffset = self.finishedOffset = - 1
                self.lastCommitTime = Utils.current_time_mills()

        self.logger.warning("process unexcepted fetchException:" + str(traceback.format_exc()))

    #
    # query start offset to read, if failed, throw the exception
    #
    @abc.abstractmethod
    def init_start_offset(self):
        pass

    #
    # commit the last processed offset and update the startOffset
    # throw the Exception when appear error
    #
    @abc.abstractmethod
    def commit_check_point(self):
        pass

    #
    # you should implement this method as follow process:
    # 1.control fetch qps by fetchInterval
    # 2.fetchMessage with try / catch structure and process exception
    # 2.1 catch chunk process PARTITION_NOT_SERVING by sleep a while
    # 2.2 catch chunk process MESSAGE_OFFSET_OUT_OF_RANGE by fixOffsetOutofRange()
    # 2.3 reset lastFetchTime
    # 3.process fetched message by MessageProcessor and update finishedOffset / startOffset
    # 4.check whether should commit offset
    #
    @abc.abstractmethod
    def fetch_data(self):
        pass




# -*- coding:utf-8 -*-
#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 


from talos.thrift.message.ttypes import MessageCompressionType
from talos.thrift.message.ttypes import MessageBlock
from talos.thrift.message.ttypes import MessageAndOffset
from talos.thrift.common.ttypes import GalaxyTalosException
from talos.client.serialization.MessageSerializationFactory import MessageSerializationFactory
from talos.utils.Utils import current_time_mills
from talos.client.serialization.MessageSerialization import MessageSerialization
from talos.client.Constants import Constants
from talos.client.compression.Gzip import Gzip
from talos.client.compression.Snappy import Snappy
from io import BytesIO
import logging
import traceback


class Compression(object):

    logger = logging.getLogger("Compression")
    gzip = Gzip()
    snappy = Snappy()

    def compress(self, msgList=None, compressionType=None):
        return self.do_compress(msgList, compressionType, MessageSerializationFactory().get_default_message_version())

    def do_compress(self, msgList=None, compressionType=None, messageVersion=None):
        messageBlock = MessageBlock(startMessageOffset=0, messageNumber=0,
                                    messageBlockSize=0, appendTimestamp=0,
                                    createTimestamp=0, transactionId=0)
        messageCompressionType = MessageCompressionType._NAMES_TO_VALUES[compressionType]
        messageBlock.compressionType = messageCompressionType
        messageBlock.messageNumber = len(msgList)

        createTime = current_time_mills()
        if len(msgList) > 0:
            createTime = msgList[0].createTimestamp

        try:
            # assume the message bytes size is 256KB
            messageBlockData = bytearray()
            messageSerializedBuffer = bytearray()
            dataOutputIo = BytesIO(messageSerializedBuffer)
            index = 0
            for message in msgList:
                MessageSerialization().serialize_message(message, dataOutputIo, messageVersion)
                createTime += (message.createTimestamp - createTime) / (index + 1)
            messageBlock.createTimestamp = int(createTime)

            if messageCompressionType == MessageCompressionType.NONE:
                messageBlockData = dataOutputIo.getvalue()
            elif messageCompressionType == MessageCompressionType.GZIP:
                messageBlockData = self.gzip.compress(dataOutputIo.getvalue())
            elif messageCompressionType == MessageCompressionType.SNAPPY:
                messageBlockData = self.snappy.encode(dataOutputIo.getvalue())
            else:
                raise RuntimeError("Unsupported Compression Type: " + str(compressionType))

            # close dataOutputIo
            dataOutputIo.close()

            if len(messageBlockData) > Constants.TALOS_MESSAGE_BLOCK_BYTES_MAXIMAL:
                raise GalaxyTalosException(errMsg="MessageBlock must be less than or equal to " +
                                                  Constants.TALOS_MESSAGE_BLOCK_BYTES_MAXIMAL +
                                                  " bytes, got bytes: " + str(len(messageBlockData)))
            messageBlock.messageBlock = messageBlockData
            messageBlock.messageBlockSize = len(messageBlockData)
        except (RuntimeError, Exception) as e:
            self.logger.info("compress MessageList failed!" + str(traceback.format_exc()))
            raise e

        return messageBlock

    def decompress(self, msgBlockList=None, unHandledMessageNumber=None):
        if len(msgBlockList) > 0:
            messageAndOffsetList = []
            unHandledNumber = unHandledMessageNumber
            for messageBlock in reversed(msgBlockList):
                msgAndOffsetList = self.do_decompress(messageBlock, unHandledNumber)
                unHandledNumber += len(msgAndOffsetList)
                messageAndOffsetList[0:0] = msgAndOffsetList
            return messageAndOffsetList
        else:
            return []

    def do_decompress(self, messageBlock=None, unHandledNumber=None):
        messageBlockData = BytesIO()
        if messageBlock.compressionType == MessageCompressionType.NONE:
            messageBlockData = BytesIO(messageBlock.messageBlock)
        elif messageBlock.compressionType == MessageCompressionType.GZIP:
            messageBlockData = BytesIO(self.gzip.decompress(messageBlock.messageBlock))
        elif messageBlock.compressionType == MessageCompressionType.SNAPPY:
            messageBlockData = BytesIO(self.snappy.decode(messageBlock.messageBlock))

        messageNumber = messageBlock.messageNumber

        messageAndOffsetList = []
        try:
            i = 0
            while i < messageNumber:
                messageAndOffset = MessageAndOffset()
                messageAndOffset.messageOffset = messageBlock.startMessageOffset + i
                message = MessageSerialization().deserialize_message(messageBlockData)
                if messageBlock.appendTimestamp:
                    message.appendTimestamp = messageBlock.appendTimestamp
                messageAndOffset.message = message
                messageAndOffset.unHandledMessageNumber = unHandledNumber + messageNumber - 1 - i

                # add message to messageList
                messageAndOffsetList.append(messageAndOffset)
                i += 1
        except Exception as e:
            self.logger.error("Decompress messageBlock failed" + str(traceback.format_exc()))
        return messageAndOffsetList





#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from talos.thrift.Thrift import TType, TMessageType, TException, TApplicationException

from talos.thrift.transport import TTransport
from talos.thrift.protocol import TBinaryProtocol, TProtocol
try:
  from talos.thrift.protocol import fastbinary
except:
  fastbinary = None



class BaseQuota:
  """
  BaseQuota is the base struct of UserQuota and TopicQuota
  the quota service is made for three aspects:
  1. the resource number taken up by users
  2. the read/write qps quota
  3. the traffic qps quota


  Attributes:
   - spaceQuota: The max space can be used for specified developer/topic
  default 1T for developer, and 500G for topic

   - usedSpace: The used space quota

   - remainingSpace: The remaining space quota

   - putMessageQpsQuota: The max total qps of putMessage of specified developer/topic
  default 40000 for developer, and 4000 for topic

   - getMessageQpsQuota: The max total qps of getMessage of specified developer/topic
  default 40000 for developer, and 4000 for topic

   - putMessageBytesQpsQuota: The max total bytes of putMessage of specified developer/topic
  default 500MB for developer, and 50MB for topic

   - getMessageBytesQpsQuota: The max total bytes of getMessage of specified developer/topic
  default 500MB for developer, and 50MB for topic

  """

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'spaceQuota', None, None, ), # 1
    (2, TType.I64, 'usedSpace', None, None, ), # 2
    (3, TType.I64, 'remainingSpace', None, None, ), # 3
    (4, TType.I32, 'putMessageQpsQuota', None, None, ), # 4
    (5, TType.I32, 'getMessageQpsQuota', None, None, ), # 5
    (6, TType.I64, 'putMessageBytesQpsQuota', None, None, ), # 6
    (7, TType.I64, 'getMessageBytesQpsQuota', None, None, ), # 7
  )

  def __init__(self, spaceQuota=None, usedSpace=None, remainingSpace=None, putMessageQpsQuota=None, getMessageQpsQuota=None, putMessageBytesQpsQuota=None, getMessageBytesQpsQuota=None,):
    self.spaceQuota = spaceQuota
    self.usedSpace = usedSpace
    self.remainingSpace = remainingSpace
    self.putMessageQpsQuota = putMessageQpsQuota
    self.getMessageQpsQuota = getMessageQpsQuota
    self.putMessageBytesQpsQuota = putMessageBytesQpsQuota
    self.getMessageBytesQpsQuota = getMessageBytesQpsQuota

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.spaceQuota = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.usedSpace = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I64:
          self.remainingSpace = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I32:
          self.putMessageQpsQuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.I32:
          self.getMessageQpsQuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.I64:
          self.putMessageBytesQpsQuota = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 7:
        if ftype == TType.I64:
          self.getMessageBytesQpsQuota = iprot.readI64()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('BaseQuota')
    if self.spaceQuota is not None:
      oprot.writeFieldBegin('spaceQuota', TType.I64, 1)
      oprot.writeI64(self.spaceQuota)
      oprot.writeFieldEnd()
    if self.usedSpace is not None:
      oprot.writeFieldBegin('usedSpace', TType.I64, 2)
      oprot.writeI64(self.usedSpace)
      oprot.writeFieldEnd()
    if self.remainingSpace is not None:
      oprot.writeFieldBegin('remainingSpace', TType.I64, 3)
      oprot.writeI64(self.remainingSpace)
      oprot.writeFieldEnd()
    if self.putMessageQpsQuota is not None:
      oprot.writeFieldBegin('putMessageQpsQuota', TType.I32, 4)
      oprot.writeI32(self.putMessageQpsQuota)
      oprot.writeFieldEnd()
    if self.getMessageQpsQuota is not None:
      oprot.writeFieldBegin('getMessageQpsQuota', TType.I32, 5)
      oprot.writeI32(self.getMessageQpsQuota)
      oprot.writeFieldEnd()
    if self.putMessageBytesQpsQuota is not None:
      oprot.writeFieldBegin('putMessageBytesQpsQuota', TType.I64, 6)
      oprot.writeI64(self.putMessageBytesQpsQuota)
      oprot.writeFieldEnd()
    if self.getMessageBytesQpsQuota is not None:
      oprot.writeFieldBegin('getMessageBytesQpsQuota', TType.I64, 7)
      oprot.writeI64(self.getMessageBytesQpsQuota)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.spaceQuota is None:
      raise TProtocol.TProtocolException(message='Required field spaceQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.spaceQuota)
    value = (value * 31) ^ hash(self.usedSpace)
    value = (value * 31) ^ hash(self.remainingSpace)
    value = (value * 31) ^ hash(self.putMessageQpsQuota)
    value = (value * 31) ^ hash(self.getMessageQpsQuota)
    value = (value * 31) ^ hash(self.putMessageBytesQpsQuota)
    value = (value * 31) ^ hash(self.getMessageBytesQpsQuota)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class UserQuota:
  """
  Attributes:
   - baseQuota: The base quota member

   - topicNumberQuota: The max topic number allowed to create for specified developerId,
  default 20 (1 ~ 100);

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'baseQuota', (BaseQuota, BaseQuota.thrift_spec), None, ), # 1
    (2, TType.I32, 'topicNumberQuota', None, None, ), # 2
  )

  def __init__(self, baseQuota=None, topicNumberQuota=None,):
    self.baseQuota = baseQuota
    self.topicNumberQuota = topicNumberQuota

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.baseQuota = BaseQuota()
          self.baseQuota.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.topicNumberQuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('UserQuota')
    if self.baseQuota is not None:
      oprot.writeFieldBegin('baseQuota', TType.STRUCT, 1)
      self.baseQuota.write(oprot)
      oprot.writeFieldEnd()
    if self.topicNumberQuota is not None:
      oprot.writeFieldBegin('topicNumberQuota', TType.I32, 2)
      oprot.writeI32(self.topicNumberQuota)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.baseQuota is None:
      raise TProtocol.TProtocolException(message='Required field baseQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.baseQuota)
    value = (value * 31) ^ hash(self.topicNumberQuota)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class SetUserQuotaRequest:
  """
  Attributes:
   - developerId: The developerId to set quota;

   - userQuota: The user quota;

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'developerId', None, None, ), # 1
    (2, TType.STRUCT, 'userQuota', (UserQuota, UserQuota.thrift_spec), None, ), # 2
  )

  def __init__(self, developerId=None, userQuota=None,):
    self.developerId = developerId
    self.userQuota = userQuota

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.developerId = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRUCT:
          self.userQuota = UserQuota()
          self.userQuota.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('SetUserQuotaRequest')
    if self.developerId is not None:
      oprot.writeFieldBegin('developerId', TType.STRING, 1)
      oprot.writeString(self.developerId)
      oprot.writeFieldEnd()
    if self.userQuota is not None:
      oprot.writeFieldBegin('userQuota', TType.STRUCT, 2)
      self.userQuota.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.developerId is None:
      raise TProtocol.TProtocolException(message='Required field developerId is unset!')
    if self.userQuota is None:
      raise TProtocol.TProtocolException(message='Required field userQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.developerId)
    value = (value * 31) ^ hash(self.userQuota)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class QueryUserQuotaResponse:
  """
  Attributes:
   - userQuota: The user quota;

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'userQuota', (UserQuota, UserQuota.thrift_spec), None, ), # 1
  )

  def __init__(self, userQuota=None,):
    self.userQuota = userQuota

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.userQuota = UserQuota()
          self.userQuota.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('QueryUserQuotaResponse')
    if self.userQuota is not None:
      oprot.writeFieldBegin('userQuota', TType.STRUCT, 1)
      self.userQuota.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.userQuota is None:
      raise TProtocol.TProtocolException(message='Required field userQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.userQuota)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ListUserQuotaResponse:
  """
  Attributes:
   - userQuotaList: The developer -> UserQuota map;

  """

  thrift_spec = (
    None, # 0
    (1, TType.MAP, 'userQuotaList', (TType.STRING,None,TType.STRUCT,(UserQuota, UserQuota.thrift_spec)), None, ), # 1
  )

  def __init__(self, userQuotaList=None,):
    self.userQuotaList = userQuotaList

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.MAP:
          self.userQuotaList = {}
          (_ktype1, _vtype2, _size0 ) = iprot.readMapBegin()
          for _i4 in range(_size0):
            _key5 = iprot.readString()
            _val6 = UserQuota()
            _val6.read(iprot)
            self.userQuotaList[_key5] = _val6
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ListUserQuotaResponse')
    if self.userQuotaList is not None:
      oprot.writeFieldBegin('userQuotaList', TType.MAP, 1)
      oprot.writeMapBegin(TType.STRING, TType.STRUCT, len(self.userQuotaList))
      for kiter7,viter8 in self.userQuotaList.items():
        oprot.writeString(kiter7)
        viter8.write(oprot)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.userQuotaList is None:
      raise TProtocol.TProtocolException(message='Required field userQuotaList is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.userQuotaList)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class DeleteUserQuotaRequest:
  """
  Attributes:
   - developerId: The developerId to delete quota;

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'developerId', None, None, ), # 1
  )

  def __init__(self, developerId=None,):
    self.developerId = developerId

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.developerId = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('DeleteUserQuotaRequest')
    if self.developerId is not None:
      oprot.writeFieldBegin('developerId', TType.STRING, 1)
      oprot.writeString(self.developerId)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.developerId is None:
      raise TProtocol.TProtocolException(message='Required field developerId is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.developerId)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class PartitionQuotaInfo:
  """
  Attributes:
   - orgId: The Org Id

   - totalQuota: The org apply/old total partition quota, depending on specific request

   - usedOuota: The org used partition quota, optional

   - applyReason: The reason of apply quota

   - orgName: The Org Name

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'orgId', None, None, ), # 1
    (2, TType.I32, 'totalQuota', None, None, ), # 2
    (3, TType.I32, 'usedOuota', None, None, ), # 3
    (4, TType.STRING, 'applyReason', None, None, ), # 4
    (5, TType.STRING, 'orgName', None, None, ), # 5
  )

  def __init__(self, orgId=None, totalQuota=None, usedOuota=None, applyReason=None, orgName=None,):
    self.orgId = orgId
    self.totalQuota = totalQuota
    self.usedOuota = usedOuota
    self.applyReason = applyReason
    self.orgName = orgName

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.orgId = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.totalQuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.usedOuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRING:
          self.applyReason = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.STRING:
          self.orgName = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('PartitionQuotaInfo')
    if self.orgId is not None:
      oprot.writeFieldBegin('orgId', TType.STRING, 1)
      oprot.writeString(self.orgId)
      oprot.writeFieldEnd()
    if self.totalQuota is not None:
      oprot.writeFieldBegin('totalQuota', TType.I32, 2)
      oprot.writeI32(self.totalQuota)
      oprot.writeFieldEnd()
    if self.usedOuota is not None:
      oprot.writeFieldBegin('usedOuota', TType.I32, 3)
      oprot.writeI32(self.usedOuota)
      oprot.writeFieldEnd()
    if self.applyReason is not None:
      oprot.writeFieldBegin('applyReason', TType.STRING, 4)
      oprot.writeString(self.applyReason)
      oprot.writeFieldEnd()
    if self.orgName is not None:
      oprot.writeFieldBegin('orgName', TType.STRING, 5)
      oprot.writeString(self.orgName)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.orgId is None:
      raise TProtocol.TProtocolException(message='Required field orgId is unset!')
    if self.totalQuota is None:
      raise TProtocol.TProtocolException(message='Required field totalQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.orgId)
    value = (value * 31) ^ hash(self.totalQuota)
    value = (value * 31) ^ hash(self.usedOuota)
    value = (value * 31) ^ hash(self.applyReason)
    value = (value * 31) ^ hash(self.orgName)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ApplyQuotaRequest:
  """
  Attributes:
   - quotaInfo: partition quota info

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'quotaInfo', (PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec), None, ), # 1
  )

  def __init__(self, quotaInfo=None,):
    self.quotaInfo = quotaInfo

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.quotaInfo = PartitionQuotaInfo()
          self.quotaInfo.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ApplyQuotaRequest')
    if self.quotaInfo is not None:
      oprot.writeFieldBegin('quotaInfo', TType.STRUCT, 1)
      self.quotaInfo.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.quotaInfo is None:
      raise TProtocol.TProtocolException(message='Required field quotaInfo is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.quotaInfo)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class AutoApplyQuotaRequest:
  """
  Attributes:
   - quotaInfo: partition quota info

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'quotaInfo', (PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec), None, ), # 1
  )

  def __init__(self, quotaInfo=None,):
    self.quotaInfo = quotaInfo

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.quotaInfo = PartitionQuotaInfo()
          self.quotaInfo.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('AutoApplyQuotaRequest')
    if self.quotaInfo is not None:
      oprot.writeFieldBegin('quotaInfo', TType.STRUCT, 1)
      self.quotaInfo.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.quotaInfo is None:
      raise TProtocol.TProtocolException(message='Required field quotaInfo is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.quotaInfo)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ListQuotaResponse:
  """
  Attributes:
   - quotaList: user existing approved quota

   - approvedTotalQuota: approved quota total number

  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'quotaList', (TType.STRUCT,(PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec)), None, ), # 1
    (2, TType.I32, 'approvedTotalQuota', None, None, ), # 2
  )

  def __init__(self, quotaList=None, approvedTotalQuota=None,):
    self.quotaList = quotaList
    self.approvedTotalQuota = approvedTotalQuota

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.quotaList = []
          (_etype12, _size9) = iprot.readListBegin()
          for _i13 in range(_size9):
            _elem14 = PartitionQuotaInfo()
            _elem14.read(iprot)
            self.quotaList.append(_elem14)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.approvedTotalQuota = iprot.readI32()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ListQuotaResponse')
    if self.quotaList is not None:
      oprot.writeFieldBegin('quotaList', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.quotaList))
      for iter15 in self.quotaList:
        iter15.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.approvedTotalQuota is not None:
      oprot.writeFieldBegin('approvedTotalQuota', TType.I32, 2)
      oprot.writeI32(self.approvedTotalQuota)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.quotaList is None:
      raise TProtocol.TProtocolException(message='Required field quotaList is unset!')
    if self.approvedTotalQuota is None:
      raise TProtocol.TProtocolException(message='Required field approvedTotalQuota is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.quotaList)
    value = (value * 31) ^ hash(self.approvedTotalQuota)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ListPendingQuotaResponse:
  """
  Attributes:
   - pendingQuotaList: pending quota application

   - talosAdmin: is or not Talos Admin

  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'pendingQuotaList', (TType.STRUCT,(PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec)), None, ), # 1
    (2, TType.BOOL, 'talosAdmin', None, None, ), # 2
  )

  def __init__(self, pendingQuotaList=None, talosAdmin=None,):
    self.pendingQuotaList = pendingQuotaList
    self.talosAdmin = talosAdmin

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.pendingQuotaList = []
          (_etype19, _size16) = iprot.readListBegin()
          for _i20 in range(_size16):
            _elem21 = PartitionQuotaInfo()
            _elem21.read(iprot)
            self.pendingQuotaList.append(_elem21)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.BOOL:
          self.talosAdmin = iprot.readBool()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ListPendingQuotaResponse')
    if self.pendingQuotaList is not None:
      oprot.writeFieldBegin('pendingQuotaList', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.pendingQuotaList))
      for iter22 in self.pendingQuotaList:
        iter22.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.talosAdmin is not None:
      oprot.writeFieldBegin('talosAdmin', TType.BOOL, 2)
      oprot.writeBool(self.talosAdmin)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.pendingQuotaList is None:
      raise TProtocol.TProtocolException(message='Required field pendingQuotaList is unset!')
    if self.talosAdmin is None:
      raise TProtocol.TProtocolException(message='Required field talosAdmin is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.pendingQuotaList)
    value = (value * 31) ^ hash(self.talosAdmin)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ApproveQuotaRequest:
  """
  Attributes:
   - quotaInfo: partition quota info

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'quotaInfo', (PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec), None, ), # 1
  )

  def __init__(self, quotaInfo=None,):
    self.quotaInfo = quotaInfo

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.quotaInfo = PartitionQuotaInfo()
          self.quotaInfo.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ApproveQuotaRequest')
    if self.quotaInfo is not None:
      oprot.writeFieldBegin('quotaInfo', TType.STRUCT, 1)
      self.quotaInfo.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.quotaInfo is None:
      raise TProtocol.TProtocolException(message='Required field quotaInfo is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.quotaInfo)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ApproveQuotaResponse:
  """
  Attributes:
   - pendingQuotaList: pending quota application

   - talosAdmin: is or not Talos Admin

  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'pendingQuotaList', (TType.STRUCT,(PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec)), None, ), # 1
    (2, TType.BOOL, 'talosAdmin', None, None, ), # 2
  )

  def __init__(self, pendingQuotaList=None, talosAdmin=None,):
    self.pendingQuotaList = pendingQuotaList
    self.talosAdmin = talosAdmin

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.pendingQuotaList = []
          (_etype26, _size23) = iprot.readListBegin()
          for _i27 in range(_size23):
            _elem28 = PartitionQuotaInfo()
            _elem28.read(iprot)
            self.pendingQuotaList.append(_elem28)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.BOOL:
          self.talosAdmin = iprot.readBool()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ApproveQuotaResponse')
    if self.pendingQuotaList is not None:
      oprot.writeFieldBegin('pendingQuotaList', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.pendingQuotaList))
      for iter29 in self.pendingQuotaList:
        iter29.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.talosAdmin is not None:
      oprot.writeFieldBegin('talosAdmin', TType.BOOL, 2)
      oprot.writeBool(self.talosAdmin)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.pendingQuotaList is None:
      raise TProtocol.TProtocolException(message='Required field pendingQuotaList is unset!')
    if self.talosAdmin is None:
      raise TProtocol.TProtocolException(message='Required field talosAdmin is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.pendingQuotaList)
    value = (value * 31) ^ hash(self.talosAdmin)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class RevokeQuotaRequest:
  """
  Attributes:
   - quotaInfo: partition quota info

  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'quotaInfo', (PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec), None, ), # 1
  )

  def __init__(self, quotaInfo=None,):
    self.quotaInfo = quotaInfo

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.quotaInfo = PartitionQuotaInfo()
          self.quotaInfo.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('RevokeQuotaRequest')
    if self.quotaInfo is not None:
      oprot.writeFieldBegin('quotaInfo', TType.STRUCT, 1)
      self.quotaInfo.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.quotaInfo is None:
      raise TProtocol.TProtocolException(message='Required field quotaInfo is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.quotaInfo)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class RevokeQuotaResponse:
  """
  Attributes:
   - pendingQuotaList: pending quota application

   - talosAdmin: is or not Talos Admin

  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'pendingQuotaList', (TType.STRUCT,(PartitionQuotaInfo, PartitionQuotaInfo.thrift_spec)), None, ), # 1
    (2, TType.BOOL, 'talosAdmin', None, None, ), # 2
  )

  def __init__(self, pendingQuotaList=None, talosAdmin=None,):
    self.pendingQuotaList = pendingQuotaList
    self.talosAdmin = talosAdmin

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.pendingQuotaList = []
          (_etype33, _size30) = iprot.readListBegin()
          for _i34 in range(_size30):
            _elem35 = PartitionQuotaInfo()
            _elem35.read(iprot)
            self.pendingQuotaList.append(_elem35)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.BOOL:
          self.talosAdmin = iprot.readBool()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('RevokeQuotaResponse')
    if self.pendingQuotaList is not None:
      oprot.writeFieldBegin('pendingQuotaList', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.pendingQuotaList))
      for iter36 in self.pendingQuotaList:
        iter36.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.talosAdmin is not None:
      oprot.writeFieldBegin('talosAdmin', TType.BOOL, 2)
      oprot.writeBool(self.talosAdmin)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.pendingQuotaList is None:
      raise TProtocol.TProtocolException(message='Required field pendingQuotaList is unset!')
    if self.talosAdmin is None:
      raise TProtocol.TProtocolException(message='Required field talosAdmin is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.pendingQuotaList)
    value = (value * 31) ^ hash(self.talosAdmin)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.thrift.auth.ttypes import Credential
from talos.thrift.auth import constants
from talos.thrift.transport.TTransport import TTransportBase
from io import BytesIO
from datetime import datetime
from six.moves import urllib
from six.moves import http_client
import six
import time
import hashlib
import hmac
import base64
import os
import ssl
import sys
import uuid
import logging
import traceback


class TalosHttpClientTransportFactory:

    _url = str
    _credential = Credential
    _client = http_client.HTTPConnection
    _clockOffset = int
    _agent = str

    def __init__(self, url=None, credential=None, client=None, agent=None):
        self._url = url
        self._credential = credential
        self._client = client
        self._agent = agent

    def get_transport(self, trans=None):
        return self.get_transport_with_clock_offset(trans, 0, "")

    def get_transport_with_clock_offset(self, trans=None, clockOffset=None, query=None):
        if trans:
            pass
        t2 = TalosHttpClient(uri_or_host=self._url, credential=self._credential, clockOffset=self._clockOffset, query=query)
        return t2


class TalosHttpClient(TTransportBase):
    logger = logging.getLogger("TalosHttpClient")
    GMT_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'
    _REQUEST_ID_LENGTH = 8
    _url_ = None
    _requestBuffer_ = BytesIO
    _inputStream_ = BytesIO()
    _connectTimeout_ = 0
    _socketTimeout_ = 0
    _clockOffset = 0
    _queryString = None
    _supportAccountKey = False
    credential = Credential

    __http = None
    scheme = None
    using_proxy = None
    host = str
    port = None
    __timeout = None
    keyfile = None
    certfile = None
    context = None
    realhost = None
    realport = None
    proxy_auth = None
    path = None
    __http_response = None
    __wbuf = None
    __custom_headers = None
    __auth_headers = None
    code = None
    message = None
    _headers = dict()
    _uri_or_host = None
    _sid = None

    # def __init__(self, uri_or_host, credential: Credential, httpClient: http.client.HTTPConnection,
    #              agent: str, clockOffset: int, query: str):
    #     self._queryString = query
    #     self.__http = httpClient
    #     self.__init__(uri_or_host, credential, clockOffset, True)

    def __init__(self, uri_or_host, credential=None, clockOffset=None, supportAccountKey=None,
                port=None, path=None, cafile=None, cert_file=None, key_file=None, ssl_context=None):
        """THttpClient supports two different types of construction:

        THttpClient(host, port, path) - deprecated
        THttpClient(uri, [port=<n>, path=<s>, cafile=<filename>, cert_file=<filename>, key_file=<filename>, ssl_context=<context>])

        Only the second supports https.  To properly authenticate against the server,
        provide the client's identity by specifying cert_file and key_file.  To properly
        authenticate the server, specify either cafile or ssl_context with a CA defined.
        NOTE: if both cafile and ssl_context are defined, ssl_context will override cafile.
        """
        if port is not None:
            self.logger.warning(
                "Please use the THttpClient('http{s}://host:port/path') constructor",
                DeprecationWarning,
                stacklevel=2)
            self.host = uri_or_host
            self.port = port
            assert path
            self.path = path
            self.scheme = 'http'
        else:
            parsed = urllib.parse.urlparse(uri_or_host)
            self.scheme = parsed.scheme
            assert self.scheme in ('http', 'https')
            if self.scheme == 'http':
                self.port = parsed.port or http_client.HTTPConnection.default_port
            elif self.scheme == 'https':
                self.port = parsed.port or http_client.HTTPConnection.default_port
                self.certfile = cert_file
                self.keyfile = key_file
                self.context = ssl.create_default_context(cafile=cafile) if (
                            cafile and not ssl_context) else ssl_context
            self.host = parsed.hostname
            self.path = parsed.path
            if parsed.query:
                self.path += '?%s' % parsed.query
        try:
            proxy = urllib.request.getproxies()[self.scheme]
        except KeyError:
            proxy = None
        else:
            if urllib.request.proxy_bypass(self.host):
                proxy = None
        if proxy:
            parsed = urllib.parse.urlparse(proxy)
            self.realhost = self.host
            self.realport = self.port
            self.host = parsed.hostname
            self.port = parsed.port
            self.proxy_auth = self.basic_proxy_auth_header(parsed)
        else:
            self.realhost = self.realport = self.proxy_auth = None
        self._clockOffset = clockOffset
        self._supportAccountKey = supportAccountKey
        self.credential = credential
        self.__wbuf = BytesIO()
        self.__http_response = None
        self.__timeout = None
        self._uri_or_host = uri_or_host

    @staticmethod
    def basic_proxy_auth_header(proxy):
        if proxy is None or not proxy.username:
            return None
        ap = "%s:%s" % (urllib.parse.unquote(proxy.username),
                        urllib.parse.unquote(proxy.password))
        cr = base64.b64encode(ap).strip()
        return "Basic " + cr

    def using_proxy(self):
        return self.realhost is not None

    def set_timeout(self, ms):
        if ms is None:
            self.__timeout = None
        else:
            self.__timeout = int(ms / 1000.0)

    def set_custom_headers(self, headers):
        self.__custom_headers = headers

    def set_support_account_key(self, supportAccountKey=None):
        self._supportAccountKey = supportAccountKey

    # Set signature related headers when credential is properly set
    def set_authentication_headers(self, data=None):
        if self.credential:
            if self.credential.secretKey and self.credential.type:
                # signature is supported
                if constants.SIGNATURE_SUPPORT.get(self.credential.type):
                    # host
                    host = str(self.host)
                    host = host.split(":")[0]
                    self.__http.putheader(constants.HK_HOST, host)
                    self._headers[str(constants.HK_HOST).lower()] = host

                    # timestamp
                    timestamp = int(time.time()) + self._clockOffset
                    self.__http.putheader(constants.HK_TIMESTAMP, str(timestamp))
                    self._headers[str(constants.HK_TIMESTAMP).lower()] = str(timestamp)
                    mi_date = datetime.utcnow().strftime(self.GMT_FORMAT)
                    self.__http.putheader(constants.MI_DATE, mi_date)
                    self._headers[str(constants.MI_DATE).lower()] = mi_date

                    # content md5
                    m = hashlib.md5()
                    m.update(data)
                    data_md5 = m.hexdigest()
                    self.__http.putheader(constants.HK_CONTENT_MD5, data_md5)
                    self._headers[str(constants.HK_CONTENT_MD5).lower()] = data_md5

                    # signature
                    headers = dict()
                    for (k, v) in self._headers.items():
                        headers[str(k).lower()] = str(v)

                    # generate authString for header
                    try:
                        authType = "Galaxy-V2 "
                        if self._supportAccountKey:
                            authType = "Galaxy-V3 "
                        if self.credential.type == constants.UserType.APP_ACCESS_TOKEN:
                            authType = "OAuth "

                        # check secretKeyId and set header attached info
                        # secretKeyId format: "Service-Admin#SecretKeyId#developerId"
                        secretKeyId = self.credential.secretKeyId
                        if constants.HK_SERVICE_ADMIN in secretKeyId:
                            items = str(secretKeyId).split(constants.HK_SERVICE_MARK)
                            if not len(items) == 3:
                                raise RuntimeError("Invalid credential secretKeyId, " + \
                                                   "expected: 3, actual: " + str(len(items)))
                            #  reset secretKeyId and add header attached info
                            secretKeyId = items[0]
                            self.__http.putheader(constants.HK_ATTACHED_INFO, items[2])
                            self._headers[str(constants.HK_ATTACHED_INFO).lower()] = items[2]

                        authString = authType + secretKeyId + ":"
                        authString += self.generate_auth_header(self._headers, self.canonicalize_resource(self._uri_or_host))
                        self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                        self._headers[str(constants.HK_AUTHORIZATION).lower()] = authString
                    except Exception as e:
                        self.logger.error("fail to sign" + str(traceback.format_exc()))
            else:
                if self.credential.type == constants.UserType.APP_XIAOMI_SSO:
                    authString = "SSO " + self._sid + ":" + self.credential.secretKey + ":" + self.credential.secretKeyId
                    self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                elif self.credential.type == constants.UserType.APP_ANONYMOUS:
                    authString = "Guest " + self.credential.secretKeyId
                    self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                else:
                    raise RuntimeError("Unsupported user type: " + self.credential.type)
            return

    def generate_auth_header(self, headers=None, canonicalizeResource=None):
        stringToSign = "POST\n"
        if "content-md5" in headers:
            stringToSign += headers["content-md5"]
        stringToSign += '\n'
        if "content-type" in headers:
            stringToSign += headers["content-type"]
        stringToSign += '\n'
        stringToSign += self.canonicalize_xiaomi_headers(headers) + canonicalizeResource
        key = str(self.credential.secretKey).encode("utf-8")
        msg = stringToSign.encode("utf-8")
        mac = hmac.new(key, msg, hashlib.sha1)
        returnStr = str(base64.b64encode(mac.digest()).decode("utf-8"))
        return returnStr

    def canonicalize_xiaomi_headers(self, headers=None):
        if not headers:
            return ""
        else:
            canonicalizedKeys = list()
            canonicalizedHeaders = dict()
            for (k, v) in headers.items():
                lowerKey = str(k).lower()
                if "x-xiaomi-" in lowerKey:
                    canonicalizedKeys.append(lowerKey)
                    canonicalizedHeaders[lowerKey] = v + '\n'
            canonicalizedKeys.sort()
            result = "\n"
            for k in canonicalizedKeys:
                result += k
                result += ':'
                result += canonicalizedHeaders[k]
            return result

    def canonicalize_resource(self, uri=None):
        subResource = ["acl", "quota", "uploads", "partNumber", "uploadId", "storageAccessToken", "metadata"]
        parsedUrl = http_client.urlsplit(url=uri)
        result = parsedUrl.path
        queryArgs = parsedUrl.query
        canonicalizeQuery = []
        for k in queryArgs:
            if k in subResource:
                canonicalizeQuery.append(k)
        if not len(canonicalizeQuery) == 0:
            i = 0
            canonicalizeQuery.sort()
            for k in canonicalizeQuery:
                if i == 0:
                    result += '?'
                else:
                    result += '&'
                values = queryArgs[k]
                if len(values) == 1 and values[0] == "":
                    result += k
                else:
                    result += k
                    result += '='
                    result += values[len(values) - 1]
                i += 1

        return result

    def isOpen(self):
        return self.__http is not None

    def open(self):
        if self.scheme == 'http':
            self.__http = http_client.HTTPConnection(self.host, self.port,
                                                     timeout=self.__timeout)
        elif self.scheme == 'https':
            self.__http = http_client.HTTPSConnection(self.host, self.port,
                                                      key_file=self.keyfile,
                                                      cert_file=self.certfile,
                                                      timeout=self.__timeout,
                                                      context=self.context)
        if self.using_proxy():
            self.__http.set_tunnel(self.realhost, self.realport,
                                   {"Proxy-Authorization": self.proxy_auth})

    def close(self):
        self.__http.close()
        self.__http = None
        self.__http_response = None

    def read(self, sz):
        return self.__http_response.read(sz)

    def write(self, buf):
        self.__wbuf.write(buf)

    def flush(self):
        if self.isOpen():
            self.close()
        self.open()

        # Pull data out of buffer
        data = self.__wbuf.getvalue()
        self.__wbuf = BytesIO()

        # HTTP request
        requestId = self.generate_random_id(self._REQUEST_ID_LENGTH)
        sb = str(self.path)
        sb += "?id="
        sb += requestId
        if self._queryString:
            sb += "&"
            sb += self._queryString

        if self.using_proxy() and self.scheme == "http":
            # need full URL of real host for HTTP proxy here (HTTPS uses CONNECT tunnel)
            self.__http.putrequest('POST', "http://%s:%s%s" %
                                   (self.realhost, self.realport, sb))
        else:
            self.__http.putrequest('POST', sb)

        # Write headers
        self.__http.putheader('Content-Type', 'application/x-thrift-compact')
        self._headers['Content-Type'.lower()] = 'application/x-thrift-compact'
        self.__http.putheader('Content-Length', str(len(data)))
        self._headers['Content-Length'] = str(len(data))
        self.__http.putheader('Accept', 'application/x-thrift-compact')
        self._headers['Accept'.lower()] = 'application/x-thrift-compact'
        if self.using_proxy() and self.scheme == "http" and self.proxy_auth is not None:
            self.__http.putheader("Proxy-Authorization", self.proxy_auth)
            self._headers["Proxy-Authorization".lower()] = self.proxy_auth

        if not self.__custom_headers or 'User-Agent' not in self.__custom_headers:
            user_agent = 'Python/THttpClient'
            script = os.path.basename(sys.argv[0])
            if script:
                user_agent = '%s (%s)' % (user_agent, urllib.parse.quote(script))
            self.__http.putheader('User-Agent', user_agent)
            self._headers['User-Agent'.lower()] = user_agent

        if self.__custom_headers:
            for key, val in six.iteritems(self.__custom_headers):
                self.__http.putheader(key, val)
                self._headers[key] = val

        self.set_authentication_headers(data)

        self.__http.endheaders()

        # Write payload
        self.__http.send(data)

        # Get reply to flush the request
        self.__http_response = self.__http.getresponse()
        response = self.__http_response
        self.code = self.__http_response.status
        self.message = self.__http_response.reason
        self._headers = self.__http_response.msg
        if not self.code == 200:
            pass
        # TODO Trans Exception to TalosTransportError
        # reason = response.fp.read()
        # raise TalosTransportError(self.code, self.message, current_time_mills())

    def generate_random_id(self, length=None):
        return str(uuid.uuid4())[0:length]

    def set_query_string(self, queryString=None):
        self._queryString = queryString

    def set_clock_offset(self, clockOffset=None):
        self._clockOffset = clockOffset

    def set_sid(self, sid=None):
        self._sid = sid




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 
 

class MessageCheckpointer:
    #
    # When user call checkpoint, TalosConsumer will checkpoint the latest
    # messageOffset(We just call it commitOffset) that had been delivered to user
    # by MessageProcessor.Once failover happens, the new MessageProcessor will
    # call it's interface init() with (commitOffset + 1) as the startMessageOffset,
    # and the new MessageProcessor will deliver message with offset equals
    # (commitOffset + 1) as the first message.
    # User should call this method periodically(e.g.once every 1 minutes),
    # Calling this API too frequently can slow down the application.
    # @ return indicate checkpoint is success or not;
    #
    def check_point(self):
        pass

    #
    # Same as checkpoint, But use commotOffset as the checkpoint value.
    # There are some restrict for commitOffset:
    # 1. commitOffset must always greater or equal than startMessageOffset;
    # 2. commitOffset must always greater than last commitOffset;
    # 3. commitOffset must always littler or equal than latest deliver messageOffset.
    # @ param commitOffset the messageOffset that to checkpoint
    # @ return indicate checkpoint is success or not;
    #
    def check_point_with_offset(self, commitOffset=None):
        pass



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.client.TalosClientConfig import TalosClientConfig
from talos.thrift.auth.ttypes import Credential
from talos.thrift.common.ttypes import Version
from talos.thrift.topic import TopicService
from talos.thrift.message import MessageService
from talos.thrift.consumer import ConsumerService
from talos.thrift.quota import QuotaService
from talos.client.TalosErrors import InvalidArgumentError
from talos.client.Constants import Constants
from talos.thrift.protocol.TCompactProtocol import TCompactProtocol
from talos.client.TalosHttpClient import TalosHttpClient
import logging
import platform
import http.client
import http


class ConsumerClient:
    talosHttpClient = TalosHttpClient

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def renew(self, request=None):
        self.talosHttpClient.set_query_string("type=renew")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).renew(request)

    def lock_worker(self, request=None):
        self.talosHttpClient.set_query_string("type=lockWorker")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).lockWorker(request)

    def query_worker(self, request=None):
        self.talosHttpClient.set_query_string("type=queryWorker")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).queryWorker(request)

    def lock_partition(self, request=None):
        self.talosHttpClient.set_query_string("type=lockPartition")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).lockPartition(request)

    def unlock_partition(self, request=None):
        self.talosHttpClient.set_query_string("type=unlockPartition")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).unlockPartition(request)

    def query_offset(self, request=None):
        self.talosHttpClient.set_query_string("type=queryOffset")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).queryOffset(request)

    def update_offset(self, request=None):
        self.talosHttpClient.set_query_string("type=updateOffset")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).updateOffset(request)

    def get_worker_id(self, request=None):
        self.talosHttpClient.set_query_string("type=getWorkerId")
        iprot = TCompactProtocol(self.talosHttpClient)
        return ConsumerService.Client(iprot).getWorkerId(request).workerId


class TopicClient:
    talosHttpClient = TalosHttpClient

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def create_topic(self, request=None):
        self.talosHttpClient.set_query_string("type=createTopic")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).createTopic(request)

    def get_describe_info(self, request=None):
        self.talosHttpClient.set_query_string("type=getDescribeInfo")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).getDescribeInfo(request)

    def get_topic_attribute(self, request=None):
        self.talosHttpClient.set_query_string("type=getTopicAttribute")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).getTopicAttribute(request)

    def delete_topic(self, request=None):
        self.talosHttpClient.set_query_string("type=deleteTopic")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).deleteTopic(request)

    def change_topic_attribute(self, request=None):
        self.talosHttpClient.set_query_string("type=changeTopicAttribute")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).changeTopicAttribute(request)

    def list_topics(self):
        self.talosHttpClient.set_query_string("type=listTopics")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).listTopics().topicInfos

    def set_permission(self, request=None):
        assert request.permission > 0
        self.talosHttpClient.set_query_string("type=setPermission")
        iprot = TCompactProtocol(self.talosHttpClient)
        TopicService.Client(iprot).setPermission(request)

    def revoke_permission(self, request=None):
        self.talosHttpClient.set_query_string("type=revokePermission")
        iprot = TCompactProtocol(self.talosHttpClient)
        TopicService.Client(iprot).setPermission(request)

    def list_permission(self, request=None):
        self.talosHttpClient.set_query_string("type=listPermission")
        iprot = TCompactProtocol(self.talosHttpClient)
        return TopicService.Client(iprot).listPermission(request).permissions


class MessageClient:
    talosHttpClient = TalosHttpClient

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def get_message(self, request=None):
        self.talosHttpClient.set_query_string("type=getMessage")
        iprot = TCompactProtocol(self.talosHttpClient)
        return MessageService.Client(iprot).getMessage(request)

    def put_message(self, request=None):
        self.talosHttpClient.set_query_string("type=putMessage")
        iprot = TCompactProtocol(self.talosHttpClient)
        return MessageService.Client(iprot).putMessage(request)

    def get_topic_offset(self, request=None):
        self.talosHttpClient.set_query_string("type=getTopicOffset")
        iprot = TCompactProtocol(self.talosHttpClient)
        return MessageService.Client(iprot).getTopicOffset(request).offsetInfoList

    def get_partition_offset(self, request=None):
        self.talosHttpClient.set_query_string("type=getPartitionOffset")
        iprot = TCompactProtocol(self.talosHttpClient)
        return MessageService.Client(iprot).getPartitionOffset(request).offsetInfo

    def get_schedule_info(self, request=None):
        self.talosHttpClient.set_query_string("type=getScheduleInfo")
        iprot = TCompactProtocol(self.talosHttpClient)
        return MessageService.Client(iprot).getScheduleInfo(request).scheduleInfo


class QuotaClient:
    talosHttpClient = TalosHttpClient

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def apply_quota(self, request=None):
        self.talosHttpClient.set_query_string("type=applyQuota")
        iprot = TCompactProtocol(self.talosHttpClient)
        QuotaService.Client(iprot).applyQuota(request)

    def revoke_quota(self, request=None):
        self.talosHttpClient.set_query_string("type=revokeQuota")
        iprot = TCompactProtocol(self.talosHttpClient)
        return QuotaService.Client(iprot).revokeQuota(request)

    def list_quota(self):
        self.talosHttpClient.set_query_string("type=listQuota")
        iprot = TCompactProtocol(self.talosHttpClient)
        return QuotaService.Client(iprot).listQuota()


class TalosClientFactory:
    logger = logging.getLogger("TalosClientFactory")
    _USER_AGENT_HEADER = "User-Agent"
    _SID = "galaxytalos"
    _DEFAULT_CLIENT_CONN_TIMEOUT = 5000

    _version = Version
    _talosClientConfig = TalosClientConfig
    _credential = Credential
    _customHeaders = dict
    _httpClient = http.client.HTTPConnection
    _agent = str
    _clockOffset = int

    def __init__(self, clientConfig=None, credential=None):
        self._talosClientConfig = clientConfig
        self._credential = credential
        self._customHeaders = None
        self._version = Version()
        self._clockOffset = 0

    def new_topic_client(self):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                          + Constants.TALOS_TOPIC_SERVICE_PATH,
                                          self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return TopicClient(talosHttpClient)

    def new_consumer_client(self):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                          + Constants.TALOS_CONSUMER_SERVICE_PATH,
                                          self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return ConsumerClient(talosHttpClient)

    def set_custom_headers(self, customHeaders=None):
        self._customHeaders = customHeaders

    def new_message_client(self):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                          + Constants.TALOS_MESSAGE_SERVICE_PATH,
                                          self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return MessageClient(talosHttpClient)

    def check_credential(self):
        if not self._credential:
            raise InvalidArgumentError("Credential is not set")

    def create_user_agent_header(self):
        return "Python-SDK/" + str(self._version.major) + "." + str(self._version.minor) + "." + str(self._version.revision) + " Python/" + platform.python_version()





#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from talos.thrift.Thrift import TType, TMessageType, TException, TApplicationException

from talos.thrift.transport import TTransport
from talos.thrift.protocol import TBinaryProtocol, TProtocol

try:
    from talos.thrift.protocol import fastbinary
except:
    fastbinary = None

class HttpStatusCode:
  BAD_REQUEST = 400
  INVALID_AUTH = 401
  CLOCK_TOO_SKEWED = 412
  REQUEST_TOO_LARGE = 413
  INTERNAL_ERROR = 500

class ErrorCode:
    NO_ERROR = 0
    UNKNOWN = 1
    UNEXPECTED_ERROR = 2
    UNEXPECTED_MESSAGE_OFFSET = 3
    TOPIC_EXIST = 4
    TOPIC_NOT_EXIST = 5
    QUOTA_NOT_EXIST = 6
    LOCK_EXIST = 7
    LOCK_NOT_EXIST = 8
    LOCK_VALUE_NOT_EXPECTED = 9
    MESSAGE_MISSING = 10
    MESSAGE_INCOMPLETE = 11
    MESSAGE_INDEX_UNDESIRED_ERROR = 12
    MESSAGE_INDEX_NOT_EXIST = 13
    MESSAGE_OFFSET_OUT_OF_RANGE = 14
    INVALID_NAME_ERROR = 15
    INVALID_TOPIC_PARAMS = 16
    OPERATION_FAILED = 17
    HDFS_OPERATION_FAILED = 18
    HBASE_OPERATION_FAILED = 19
    ZOOKEEPER_OPERATION_FAILED = 20
    PARTITION_NOT_SERVING = 21
    PARTITION_NOT_EXIST = 22
    ZK_NODE_EXIST = 23
    ZK_NODE_NOT_EXIST = 24
    REST_SERVER_INIT_ERROR = 25
    INTERNAL_SERVER_ERROR = 26
    EXCESSIVE_PENDING_MESSAGE = 27
    PERMISSION_DENIED_ERROR = 28
    HDFS_FILE_NOT_EXIST = 29
    INVALID_AUTH_INFO = 30
    SUB_RESOURCE_NAME_EXIST = 31
    SUB_RESOURCE_NAME_NOT_EXIST = 32
    REQUEST_PROCESS_TIMEOUT = 33
    QUOTA_EXCEEDED = 34
    THROTTLE_REJECT_ERROR = 35
    QUOTA_AUTO_APPROVE_FAILED = 36
    HBASE_OPERATION_BLOCKED = 37
    CLOCK_TOO_SKEWED = ErrorCode = 38
    REQUEST_TOO_LARGE = 39
    BAD_REQUEST = 40

    _VALUES_TO_NAMES = {
        0: "NO_ERROR",
        1: "UNKNOWN",
        2: "UNEXPECTED_ERROR",
        3: "UNEXPECTED_MESSAGE_OFFSET",
        4: "TOPIC_EXIST",
        5: "TOPIC_NOT_EXIST",
        6: "QUOTA_NOT_EXIST",
        7: "LOCK_EXIST",
        8: "LOCK_NOT_EXIST",
        9: "LOCK_VALUE_NOT_EXPECTED",
        10: "MESSAGE_MISSING",
        11: "MESSAGE_INCOMPLETE",
        12: "MESSAGE_INDEX_UNDESIRED_ERROR",
        13: "MESSAGE_INDEX_NOT_EXIST",
        14: "MESSAGE_OFFSET_OUT_OF_RANGE",
        15: "INVALID_NAME_ERROR",
        16: "INVALID_TOPIC_PARAMS",
        17: "OPERATION_FAILED",
        18: "HDFS_OPERATION_FAILED",
        19: "HBASE_OPERATION_FAILED",
        20: "ZOOKEEPER_OPERATION_FAILED",
        21: "PARTITION_NOT_SERVING",
        22: "PARTITION_NOT_EXIST",
        23: "ZK_NODE_EXIST",
        24: "ZK_NODE_NOT_EXIST",
        25: "REST_SERVER_INIT_ERROR",
        26: "INTERNAL_SERVER_ERROR",
        27: "EXCESSIVE_PENDING_MESSAGE",
        28: "PERMISSION_DENIED_ERROR",
        29: "HDFS_FILE_NOT_EXIST",
        30: "INVALID_AUTH_INFO",
        31: "SUB_RESOURCE_NAME_EXIST",
        32: "SUB_RESOURCE_NAME_NOT_EXIST",
        33: "REQUEST_PROCESS_TIMEOUT",
        34: "QUOTA_EXCEEDED",
        35: "THROTTLE_REJECT_ERROR",
        36: "QUOTA_AUTO_APPROVE_FAILED",
        37: "HBASE_OPERATION_BLOCKED",
        38: "CLOCK_TOO_SKEWE",
        39: "REQUEST_TOO_LARGE",
        40: "BAD_REQUEST",
    }

    _NAMES_TO_VALUES = {
        "NO_ERROR": 0,
        "UNKNOWN": 1,
        "UNEXPECTED_ERROR": 2,
        "UNEXPECTED_MESSAGE_OFFSET": 3,
        "TOPIC_EXIST": 4,
        "TOPIC_NOT_EXIST": 5,
        "QUOTA_NOT_EXIST": 6,
        "LOCK_EXIST": 7,
        "LOCK_NOT_EXIST": 8,
        "LOCK_VALUE_NOT_EXPECTED": 9,
        "MESSAGE_MISSING": 10,
        "MESSAGE_INCOMPLETE": 11,
        "MESSAGE_INDEX_UNDESIRED_ERROR": 12,
        "MESSAGE_INDEX_NOT_EXIST": 13,
        "MESSAGE_OFFSET_OUT_OF_RANGE": 14,
        "INVALID_NAME_ERROR": 15,
        "INVALID_TOPIC_PARAMS": 16,
        "OPERATION_FAILED": 17,
        "HDFS_OPERATION_FAILED": 18,
        "HBASE_OPERATION_FAILED": 19,
        "ZOOKEEPER_OPERATION_FAILED": 20,
        "PARTITION_NOT_SERVING": 21,
        "PARTITION_NOT_EXIST": 22,
        "ZK_NODE_EXIST": 23,
        "ZK_NODE_NOT_EXIST": 24,
        "REST_SERVER_INIT_ERROR": 25,
        "INTERNAL_SERVER_ERROR": 26,
        "EXCESSIVE_PENDING_MESSAGE": 27,
        "PERMISSION_DENIED_ERROR": 28,
        "HDFS_FILE_NOT_EXIST": 29,
        "INVALID_AUTH_INFO": 30,
        "SUB_RESOURCE_NAME_EXIST": 31,
        "SUB_RESOURCE_NAME_NOT_EXIST": 32,
        "REQUEST_PROCESS_TIMEOUT": 33,
        "QUOTA_EXCEEDED": 34,
        "THROTTLE_REJECT_ERROR": 35,
        "QUOTA_AUTO_APPROVE_FAILED": 36,
        "HBASE_OPERATION_BLOCKED": 37,
        "CLOCK_TOO_SKEWE": 38,
        "REQUEST_TOO_LARGE": 39,
        "BAD_REQUEST": 40,
    }


class RetryType:
    SAFE = 0
    UNSAFE = 1

    _VALUES_TO_NAMES = {
        0: "SAFE",
        1: "UNSAFE",
    }

    _NAMES_TO_VALUES = {
        "SAFE": 0,
        "UNSAFE": 1,
    }


class GalaxyTalosException(TException):
    """
  UNSAFE: when user set isRetry true, will retry.

  Attributes:
   - errorCode
   - errMsg
   - details
  """

    thrift_spec = (
        None,  # 0
        (1, TType.I32, 'errorCode', None, None,),  # 1
        (2, TType.STRING, 'errMsg', None, None,),  # 2
        (3, TType.STRING, 'details', None, None,),  # 3
    )

    def __init__(self, errorCode=None, errMsg=None, details=None, ):
        self.errorCode = errorCode
        self.errMsg = errMsg
        self.details = details

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(
                iprot.trans,
                TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans,
                                     (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.errorCode = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.errMsg = iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.details = iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(
                fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('GalaxyTalosException')
        if self.errorCode is not None:
            oprot.writeFieldBegin('errorCode', TType.I32, 1)
            oprot.writeI32(self.errorCode)
            oprot.writeFieldEnd()
        if self.errMsg is not None:
            oprot.writeFieldBegin('errMsg', TType.STRING, 2)
            oprot.writeString(self.errMsg)
            oprot.writeFieldEnd()
        if self.details is not None:
            oprot.writeFieldBegin('details', TType.STRING, 3)
            oprot.writeString(self.details)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __hash__(self):
        value = 17
        value = (value * 31) ^ hash(self.errorCode)
        value = (value * 31) ^ hash(self.errMsg)
        value = (value * 31) ^ hash(self.details)
        return value

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Version:
    """
  Attributes:
   - major: The major version number;

   - minor: The minor version number;

   - revision: The revision number;

   - date: The date for release this version;

   - details: The version details;

  """

    thrift_spec = (
        None,  # 0
        (1, TType.I32, 'major', None, 1,),  # 1
        (2, TType.I32, 'minor', None, 0,),  # 2
        (3, TType.I32, 'revision', None, 0,),  # 3
        (4, TType.STRING, 'date', None, "19700101",),  # 4
        (5, TType.STRING, 'details', None, "",),  # 5
    )

    def __init__(self, major=thrift_spec[1][4], minor=thrift_spec[2][4],
                 revision=thrift_spec[3][4], date=thrift_spec[4][4],
                 details=thrift_spec[5][4], ):
        self.major = major
        self.minor = minor
        self.revision = revision
        self.date = date
        self.details = details

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(
                iprot.trans,
                TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans,
                                     (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.major = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.minor = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.revision = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.date = iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.details = iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(
                fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('Version')
        if self.major is not None:
            oprot.writeFieldBegin('major', TType.I32, 1)
            oprot.writeI32(self.major)
            oprot.writeFieldEnd()
        if self.minor is not None:
            oprot.writeFieldBegin('minor', TType.I32, 2)
            oprot.writeI32(self.minor)
            oprot.writeFieldEnd()
        if self.revision is not None:
            oprot.writeFieldBegin('revision', TType.I32, 3)
            oprot.writeI32(self.revision)
            oprot.writeFieldEnd()
        if self.date is not None:
            oprot.writeFieldBegin('date', TType.STRING, 4)
            oprot.writeString(self.date)
            oprot.writeFieldEnd()
        if self.details is not None:
            oprot.writeFieldBegin('details', TType.STRING, 5)
            oprot.writeString(self.details)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.major is None:
            raise TProtocol.TProtocolException(message='Required field major is unset!')
        if self.minor is None:
            raise TProtocol.TProtocolException(message='Required field minor is unset!')
        if self.revision is None:
            raise TProtocol.TProtocolException(
                message='Required field revision is unset!')
        if self.date is None:
            raise TProtocol.TProtocolException(message='Required field date is unset!')
        return

    def __hash__(self):
        value = 17
        value = (value * 31) ^ hash(self.major)
        value = (value * 31) ^ hash(self.minor)
        value = (value * 31) ^ hash(self.revision)
        value = (value * 31) ^ hash(self.date)
        value = (value * 31) ^ hash(self.details)
        return value

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

import sys
from talos.utils.Utils import HashCode
from talos.producer import Partitioner


class SimplePartitioner(Partitioner):

    def partition(self, partitionKey=None, partitionNum=None):
        partitionInterval = sys.maxsize / partitionNum
        return ((HashCode.HashCode(partitionKey) & 0x7FFFFFFF) / partitionInterval) % partitionNum